

 ---------------------------------------------------------------------
 --
 --                            QUARTZ
 --                            v. 1.5
 --
 --     Class & object routine library for Euphoria programming
 --
 --     Created by: Rod Jackson
 --     Copyright:  February, March, May, July, December 1999
 --
 --
 --  This library is released for public, private, and commercial use
 --  free of charge. This library may be sold, but may not be
 --  liscensed, by anyone other than the author. All unauthorized
 --  versions and/or modifications of this library may only be
 --  distributed if clearly identified as being unauthorized.
 --
 --  For additional information, contact the author at:
 --
 --     rodjackson@bigfoot.com
 --
 --------------------------------------------------------------------


constant FALSE = 0
constant TRUE = 1

constant SCREEN = 1

constant OBJ_CLASSID = 1
constant NUM_HEADER_ELEMENTS = 1

constant STORED_VERSION = 1
constant STORED_CLASS_NAME = 2
constant NUM_STORED_HEADER_ELEMENTS = 2

constant MIN_COMPATIBLE_STORED_VERSION = 1.4

global constant BASE_QCLASS = "qobject"
global constant NO_CONSTRUCTOR = -1

global constant QUARTZ_VERSION = 1.5

 -----------------------------

integer HighestClassID -- used to keep track of added classes

 -- these are for the qstring type
integer Length
object  c

 ------------------------------

 -- qobject: This defines the Quartz object type, for use when
 --    creating instantiations of a Quartz class.
global type qobject (object x)

   if (sequence (x)) then
      if (length (x) != 0) then
         if (atom(x[1])) then
            if (x[1] <= HighestClassID) then
               return TRUE
            end if
         end if
      end if
   end if

   return FALSE

end type


 -- qletter: This type defines a valid letter for a Quartz name.
type qletter (integer x)

   if ((x <= 'Z') and (x >= 'A')) then
      return TRUE
   elsif ((x <= 'z') and (x >= 'a')) then
      return TRUE
   else
      return FALSE
   end if

end type


 -- qstring: Method, member and class names in Quartz must be composed
 --    of (integer) letters and/or numbers. This type enforces that
 --    requirement.
global type qstring (object x)

   if (sequence (x)) then
      Length = length (x)

      if (Length != 0) then

          if (not (qletter (x[1]))) then
             return FALSE
          end if

          for i = 2 to Length do
             c = x[i]
             if (integer (c)) then
                if (not (qletter (c))) then
                   if (c < '0') or (c > '9') then
                      return FALSE
                   end if
                end if
             else -- not an integer
                return FALSE
             end if
          end for

          return TRUE

      end if
   end if

   return FALSE

end type


 -- integer_sequence: Function ID sequences must contain integers.
 --    This type enforces that requirement.
global type integer_sequence (object x)

   if (sequence (x)) then
      for i = 1 to length (x) do
         if (not (integer (x[i]))) then
            return FALSE
         end if
      end for
   else -- not sequence
      return FALSE
   end if

   return TRUE

end type


 -- qstring_sequence: Member and method sequences must contain
 --    valid Quartz names. This type enforces that requirement.
global type qstring_sequence (object x)

   if (sequence (x)) then
      for i = 1 to length (x) do
         if (not (qstring (x[i]))) then
            return FALSE
         end if
      end for
   else -- not sequence
      return FALSE
   end if

   return TRUE

end type

 -----------------------------

 -- the various internal tables
qstring_sequence ClassNames
qstring_sequence ClassParents
sequence         ClassMembers
sequence         ClassMethods
sequence         ClassRoutineIDs
integer_sequence ClassConstructors

 -- these are used by define_qclass
integer          ParentClassID, ParentIsBaseObject,
                 InheritedConstructor, TheConstructor
qstring_sequence InheritedMembers, InheritedMethods,
                 TotalMembers, TotalMethods
integer_sequence InheritedFunctions, TotalFunctions

 -- these are used by new_qobject
qobject TheObject, AlteredObject
integer TotalNumMembers, Constructor, nClassIndex

 -- these are used by various routines
qstring  ClassName, ObjectClass, Member
integer  ClassIndex, MethodIndex, MemberIndex, Function, NumMembers,
         CurrElement, sClassIndex, LClassIndex
object   Result
sequence ObjSeq, Members, QObj

 -----------------------------

 -- treats the given sequence like a mathematical set; the given
 -- object is only added if it is not already in the sequence
function AddToSet (sequence s, object x)

   if (find (x, s) != 0) then
      return s
   else
      return prepend (s, x)
   end if

end function

 ------------------------------



 -------------------------------------------------------------------
 -- init_quartz: This procedure is to be called before any other use
 --    of the Quartz library.
 -------------------------------------------------------------------
procedure init_quartz ()

   ClassNames        = {}
   ClassParents      = {}
   ClassMembers      = {}
   ClassMethods      = {}
   ClassRoutineIDs   = {}
   ClassConstructors = {}

   HighestClassID    = 0

end procedure -- end init_quartz()


 -------------------------------------------------------------------
 -- define_qclass: This procedure sets up a table entry for a class.
 --    It must be called for each Quartz class used in a program.
 -------------------------------------------------------------------
global procedure define_qclass (qstring ClassName,
                                qstring ParentClassName,
                                qstring_sequence MemberNames,
                                qstring_sequence MethodNames,
                                integer_sequence Functions,
                                integer ConstructorID)

   HighestClassID = HighestClassID + 1

    -- make sure the class doesn't already exist
   if ((compare (ClassName, BASE_QCLASS) != 0) and
       (find (ClassName, ClassNames) = 0)) then

       -- make sure the parent class exists (or is the base object)
      ParentClassID = find (ParentClassName, ClassNames)
      ParentIsBaseObject = (compare (ParentClassName, BASE_QCLASS) = 0)
      if (ParentIsBaseObject or (ParentClassID != 0)) then

          -- determine inherited members, memthods, etc.
         if (ParentIsBaseObject) then
            InheritedMembers     = {}
            InheritedMethods     = {}
            InheritedFunctions   = {}
            InheritedConstructor = NO_CONSTRUCTOR
         else
            InheritedMembers     = ClassMembers[ParentClassID]
            InheritedMethods     = ClassMethods[ParentClassID]
            InheritedFunctions   = ClassRoutineIDs[ParentClassID]
            InheritedConstructor = ClassConstructors[ParentClassID]
         end if

          -- add the new members to the list
         TotalMembers = InheritedMembers
         for i = 1 to length (MemberNames) do
            TotalMembers = AddToSet (TotalMembers, MemberNames[i])
         end for

          -- see if there's a new constructor method
         if (ConstructorID != NO_CONSTRUCTOR) then
            TheConstructor = ConstructorID
         else
            TheConstructor = InheritedConstructor
         end if

          -- add the methods & corresponding function IDs to the lists
         TotalMethods = InheritedMethods
         TotalFunctions = InheritedFunctions
         for i = 1 to length (MethodNames) do
            TotalMethods = prepend (TotalMethods, MethodNames[i])
            TotalFunctions = prepend (TotalFunctions, Functions[i])
         end for

          -- now we finally add the class
         ClassNames        = append (ClassNames, ClassName)
         ClassParents      = append (ClassParents, ParentClassName)
         ClassMembers      = append (ClassMembers, TotalMembers)
         ClassMethods      = append (ClassMethods, TotalMethods)
         ClassRoutineIDs   = append (ClassRoutineIDs, TotalFunctions)
         ClassConstructors = ClassConstructors & TheConstructor

      else -- the parent class doesn't exist & isn't the base object

         puts (SCREEN, "ERROR: define_qclass - " & ParentClassName &
                       " is an undefined parent class\n")
         abort (1)

      end if

   else -- class already exits

      puts (SCREEN, "ERROR: define_qclass - class " & ClassName &
                    " is already defined\n")

      abort (1)

   end if

end procedure -- end define_qclass()


 ------------------------------------------------------------------
 -- new_qobject: This function returns a Quartz object of the
 --    specified class, ready for use and manipulation via methods.
 ------------------------------------------------------------------
global function new_qobject (qstring ClassName, object Param)

    -- first, make sure the class exists & is not the base class
    -- (which cannot have objects instantiated from it directly)
   if (compare (ClassName, BASE_QCLASS) = 0) then

      puts (SCREEN, "ERROR: new_qobject - no Quartz object can " &
                    "be of class " & BASE_QCLASS & "\n")
      abort (1)

   else -- the class is not the base class

       -- let's make sure the class exists
      nClassIndex = find (ClassName, ClassNames)
      if (nClassIndex = 0) then
         puts (SCREEN, "ERROR: new_qobject - class " & ClassName &
                       " is not defined\n")
         abort (1)
      end if

   end if

    -- we begin by creating a sequence with the proper class ID and
    -- the appropriate number of members
   TotalNumMembers = length (ClassMembers[nClassIndex])

   TheObject = {nClassIndex} & repeat (0, TotalNumMembers)

    -- now, see if there's a constructor method
   Constructor = ClassConstructors[nClassIndex]
   if (Constructor != NO_CONSTRUCTOR) then

       -- there is a constructor; execute it...
      AlteredObject = call_func (Constructor, {TheObject, Param})

       -- verify the fundamental correctness of the altered object
      if (length (AlteredObject) =
          TotalNumMembers + NUM_HEADER_ELEMENTS) then
         if (integer (AlteredObject[1])) then
            if (AlteredObject[1] = TheObject[1]) then

                -- everything checks OK; return the updated object
               return AlteredObject

            end if
         end if
      end if

       -- if we get here, the object was improperly altered
      puts (SCREEN, "ERROR: new_qobject - constructor method for " &
                    "class " & ClassName & " returned an invalid " &
                    "object\n")
      abort (1)

   else

       -- no constructor; just return the object
      return TheObject

   end if

end function -- end new_qobject()


 ----------------------------------------------------------------
 -- send_qmessage: This function sends a Quartz object a message,
 --    invoking the specified method, and returning the result.
 ----------------------------------------------------------------
global function send_qmessage (qobject Target, qstring Method,
                               object Param)

    -- first extract the Quartz object's class
   ClassIndex = Target[OBJ_CLASSID]

    -- now, look for the method
   MethodIndex = find (Method, ClassMethods[ClassIndex])
   if (MethodIndex = 0) then

       -- method was not found; abort
      ClassName = ClassNames[ClassIndex]
      puts (SCREEN, "ERROR: send_qmessage - method " & Method &
                    " not defined for class " & ClassName & "\n")
      abort (1)

   else

       -- we found the method; let's execute the proper function
       -- and return the result
      Function = ClassRoutineIDs[ClassIndex][MethodIndex]
      Result = call_func (Function, {Target, Param})
      return Result

   end if

end function -- end send_qmessage()


 -----------------------------------------------------------------
 -- get_qmember: This function returns the specified member from a
 --    Quartz object.
 -----------------------------------------------------------------
global function get_qmember (qobject TheObject, qstring Member)

   ClassIndex = TheObject [OBJ_CLASSID]

   MemberIndex = find (Member, ClassMembers[ClassIndex])

   if (MemberIndex = 0) then
      ClassName = ClassNames[ClassIndex]
      puts (SCREEN, "ERROR: get_qmember - member " & Member & " is " &
                    "not defined for class " & ClassName & "\n")
      abort (1)
   end if

   return TheObject[MemberIndex + NUM_HEADER_ELEMENTS]

end function -- end get_qmember()


 --------------------------------------------------------------------
 -- set_qmember: This function sets the specified member of the given
 --    Quartz object to the specified value/object, returning the
 --    modified Quartz object.
 --------------------------------------------------------------------
global function set_qmember (qobject TheObject, qstring Member, object Param)

   ClassIndex = TheObject [OBJ_CLASSID]

   MemberIndex = find (Member, ClassMembers[ClassIndex])

   if (MemberIndex = 0) then
      ClassName = ClassNames[ClassIndex]
      puts (SCREEN, "ERROR: get_qmember - member " & Member & " is " &
                    "not defined for class " & ClassName & "\n")
      abort (1)
   end if

   TheObject[MemberIndex + NUM_HEADER_ELEMENTS] = Param

   return TheObject

end function -- end set_qmember()


 --------------------------------------------------------------------
 -- get_qobject_class: This function returns the Quartz class name of
 --    the given object.
 --------------------------------------------------------------------
global function get_qobject_class (qobject TheObject)

   return ClassNames[TheObject[OBJ_CLASSID]]

end function -- end get_qobject_class()


 -------------------------------------------------------------------
 -- is_qsubclass: This function determines if one given Quartz class
 --    is a subclass (direct or indirect) of another.
 -------------------------------------------------------------------
global function is_qsubclass (qstring Subclass, qstring Superclass)

   ClassIndex = find (Subclass, ClassNames)

   if (ClassIndex = 0) then
      puts (SCREEN, "ERROR: is_qsubclass - class " & Subclass &
                    " is not defined\n")
      abort (1)
   end if

   if (find (Superclass, ClassNames) = 0) then
      if (compare (Superclass, BASE_QCLASS) = 0) then
         return TRUE
      else
         puts (SCREEN, "ERROR: is_qsubclass - class " & Superclass &
                       " is not defined\n")
         abort (1)
      end if
   end if

    -- loop up through the hiearchy until found or not found
   while (TRUE) do

       -- get the parent class
      Subclass = ClassParents[ClassIndex]

      if (compare (Subclass, BASE_QCLASS) = 0) then
          -- the parent is the base object; no match
         return FALSE
      else
         if (compare (Subclass, Superclass) = 0) then
             -- the parent is the superclass; match!
            return TRUE
         end if
      end if

       -- not found yet; move up in the hiearchy
      ClassIndex = find (Subclass, ClassNames)

   end while

end function -- end is_qsubclass()


 -----------------------------------------------------------------
 -- is_qinstantance: This function determines if the given Quartz
 --    object is of the specified class, or if it is of a subclass
 --    (direct or indirect) of that class.
 -----------------------------------------------------------------
global function is_qinstance (qobject QObject, qstring Class)

   ClassIndex = find (Class, ClassNames)

   if (ClassIndex = 0) and (compare (Class, BASE_QCLASS) != 0) then
      puts (SCREEN, "ERROR: is_qinstance - class " & Class &
                    " is not defined\n")
      abort (1)
   end if

   ObjectClass = get_qobject_class (QObject)

    -- loop up through the hiearchy until found or not found
   while (TRUE) do

      if (compare (ObjectClass, Class) = 0) then
          -- the classes match!
         return TRUE
      else
         if (compare (ObjectClass, BASE_QCLASS) = 0) then
             -- we hit the top without a match; we're done
            return FALSE
         end if
      end if

       -- not found yet; move up in the hiearchy
      ClassIndex = find (ObjectClass, ClassNames)
      ObjectClass = ClassParents[ClassIndex]

   end while

end function -- end is_qinstance()


 -----------------------------------------------------------------
 -- store_qobject: This function converts the given Quartz object
 --    to a form more suitable for long-term storage on disk.
 --    load_qobject() may then be used to retrieve the data.
 -----------------------------------------------------------------
global function store_qobject (qobject TheObject)

   sClassIndex = TheObject[OBJ_CLASSID]
   Members     = ClassMembers[sClassIndex]
   NumMembers  = length (Members)

   ObjSeq = repeat (0, NUM_STORED_HEADER_ELEMENTS + (NumMembers * 2))

   ObjSeq[STORED_VERSION]    = MIN_COMPATIBLE_STORED_VERSION
   ObjSeq[STORED_CLASS_NAME] = ClassNames[sClassIndex]

   CurrElement = NUM_STORED_HEADER_ELEMENTS + 1
   for i = 1 to NumMembers do
      Member = Members[i]
      ObjSeq[CurrElement]     = Member
      ObjSeq[CurrElement + 1] = get_qmember (TheObject, Member)
      CurrElement = CurrElement + 2
   end for

   return ObjSeq

end function -- end store_qobject()


 -----------------------------------------------------------------
 -- load_qobject: This function returns the Quartz object that is
 --    stored in the alternate form in the sequence parameter. It
 --    is assumed that the parameter was generated by the
 --    store_qobject() function.
 -----------------------------------------------------------------
global function load_qobject (sequence StoredObject)

   if (StoredObject[STORED_VERSION] >
       MIN_COMPATIBLE_STORED_VERSION) then

      puts (SCREEN, "ERROR: load_qobject - attempt to load a stored" &
                    "q-object from an incompatible Quartz version\n")
      abort (1)

   else

      LClassIndex = find (StoredObject[STORED_CLASS_NAME], ClassNames)
      NumMembers = (length(StoredObject) - NUM_STORED_HEADER_ELEMENTS)
                   / 2

      QObj = repeat (0, NUM_HEADER_ELEMENTS + NumMembers)
      QObj[OBJ_CLASSID] = LClassIndex

      CurrElement = NUM_STORED_HEADER_ELEMENTS + 1
      for i = 1 to NumMembers do
         QObj = set_qmember (QObj, StoredObject[CurrElement],
                                   StoredObject[CurrElement + 1])
         CurrElement = CurrElement + 2
      end for

      return QObj

   end if

end function -- end load_qobject()


 ---------------------------------
 -- initialize the Quartz routines
 ---------------------------------

init_quartz ()

