
include quartz.e
without warning

 -------------------
 -- define constants
 -------------------

constant FALSE   = 0
constant TRUE    = 1
constant NIL     = 0
constant M_SOUND = 1

 ------------------------
 -- define local routines
 ------------------------

 -- borrowed from Euphoria's GRAPHICS.E file
procedure my_sound (integer f) -- f should be >= 0

   machine_proc (M_SOUND, f)

end procedure -- end my_sound()

 -----------------------
 -- define class methods
 -----------------------

global function Animal_Animal (qobject TheAnimal, object Param)

    qobject AnimalCopy
    AnimalCopy = TheAnimal

    -- Param is a sequence; the animal's name, and if domesticated
    -- if we wanted to type check Param, we should do it here, not
    -- in the function's parameter list

    -- set the animal's members to the initial values
   AnimalCopy = set_qmember (AnimalCopy, "Name", Param[1])
   AnimalCopy = set_qmember (AnimalCopy, "IsDomesticated", Param[2])

    -- return the modified animal; this is returned to "new_qobject"
   return AnimalCopy

end function -- end Animal_Animal()


global function Animal_MakeNoise (qobject TheAnimal, object Param)
    -- make a generic animal sound
   my_sound (800)
   for i = 1 to 100000 do end for
   my_sound (0)
    -- for simplicity, all functions must return a value
   return NIL
end function -- end Animal_MakeNoise()


global function Fish_MakeNoise (qobject TheFish, object Param)
    -- no sound
   my_sound (0)
   return NIL
end function -- end Fish_MakeNoise()


global function Mammal_Mammal (qobject TheMammal, object Param)

   qobject MammalCopy
   MammalCopy = TheMammal

    -- param needs to be at least 2 elements for name & domestication
   MammalCopy = set_qmember (MammalCopy, "Name", Param[1])
   MammalCopy = set_qmember (MammalCopy, "IsDomesticated", Param[2])

    -- if extra elements, apply to next members; this constructor
    -- could also have been designed to accept a fixed number of
    -- initialization elements

   if (length (Param) > 2) then
      MammalCopy = set_qmember (MammalCopy, "HasTail", Param[3])
   else
       -- set to the default
      MammalCopy = set_qmember (MammalCopy, "HasTail", TRUE)
   end if

   if (length (Param) > 3) then
      MammalCopy = set_qmember (MammalCopy, "HasFur", Param[4])
   else
      MammalCopy = set_qmember (MammalCopy, "HasFur", TRUE)
   end if

   if (length (Param) > 4) then
      MammalCopy = set_qmember (MammalCopy, "IsAquatic", Param[5])
   else
      MammalCopy = set_qmember (MammalCopy, "IsAquatic", FALSE)
   end if

   return MammalCopy

end function -- end Mammal_Mammal()


global function Dog_Dog (qobject TheDog, object Param)
    -- do nothing, just show we can block out a constructor
   return TheDog
end function -- end Dog_Dog()


 -----------------
 -- define classes
 -----------------

define_qclass ("Animal", -- class name
               BASE_QCLASS, -- parent class
               {"Name", "IsDomesticated"}, -- member names
               {"MakeNoise"}, -- method names
               {routine_id ("Animal_MakeNoise")}, -- function IDs
               routine_id ("Animal_Animal")) -- constructor function ID

 -- note that for birds and fish, we specify no constructor, but
 -- since Animal already has one, that will still be used.

define_qclass ("Bird", "Animal",
               {"Wingspan", "MaxSpeed"},
               {}, {}, NO_CONSTRUCTOR)

define_qclass ("Fish", "Animal",
               {},
               {"MakeNoise"}, -- let's override the parent method;
                              -- fish don't make much noise!
               {routine_id ("Fish_MakeNoise")},
               NO_CONSTRUCTOR)

define_qclass ("Mammal", "Animal",
               {"HasTail", "HasFur", "IsAquatic"}, {}, {},
               routine_id ("Mammal_Mammal")) -- override constructor

define_qclass ("Cat", "Mammal", {"IsDeclawed"}, {}, {}, NO_CONSTRUCTOR)

define_qclass ("Dog", "Mammal", {}, {}, {},
               routine_id ("Dog_Dog")) -- we want a do-nothing
                                       -- constructor for dogs

 ----------------
