
 -- Speed test of Quartz routines set_qmember() & get_qmember().
 -- Tests an absolute knowledge of the element number versus a
 -- two-step inheritance. Quartz is slower, but not as much as
 -- might be assumed, and it's only noticeable at high numbers of
 -- iterations.

without type_check

include qanimal.e

constant TRUE = 1
constant FALSE = 0
constant NUM_ITERATIONS = 1000000
constant SCREEN = 1

integer  Junk
sequence G
atom     LoopOverhead, SeqAccess, QAccess
qobject  MyCat

 -- set up variables for speed test
Junk = TRUE -- variable, not a constant; emulates realistic Quartz use
MyCat = new_qobject ("Cat", {"Lucky", FALSE, FALSE, TRUE})

LoopOverhead = time ()
for i = 1 to NUM_ITERATIONS do end for
LoopOverhead = time () - LoopOverhead

 -- first we compare set_qmember to sequence element assignment
puts (SCREEN, "Please press enter...")
G = gets (0) -- delay program until user presses "Enter"

puts (SCREEN, "\nEntering sequence loop...\n")

SeqAccess = time ()
for i = 1 to NUM_ITERATIONS do
    MyCat[3] = Junk
end for
SeqAccess = time () - SeqAccess - LoopOverhead

puts (SCREEN, "Entering Quartz object loop...\n")
QAccess = time ()
for i = 1 to NUM_ITERATIONS do
   MyCat = set_qmember (MyCat, "IsDomesticated", Junk)
end for
QAccess = time () - QAccess - LoopOverhead

puts (SCREEN, "Done comparing []= to set_qmember!\n\n")
puts (SCREEN, "Sequence access (less loop overhead): ") ? SeqAccess
puts (SCREEN, "Quartz access (less loop overhead):   ") ? QAccess

 -- next we compare get_qmember to sequence element extraction
puts (SCREEN, "\n\nPlease press enter...")
G = gets (0) -- delay program until user presses "Enter"

puts (SCREEN, "\nEntering sequence loop...\n")

SeqAccess = time ()
for i = 1 to NUM_ITERATIONS do
    Junk = MyCat[3]
end for
SeqAccess = time () - SeqAccess - LoopOverhead

puts (SCREEN, "Entering Quartz object loop...\n")
QAccess = time ()
for i = 1 to NUM_ITERATIONS do
   Junk = get_qmember (MyCat, "IsDomesticated")
end for
QAccess = time () - QAccess - LoopOverhead

puts (SCREEN, "Done comparing =[] to get_qmember!\n\n")
puts (SCREEN, "Sequence access (less loop overhead): ") ? SeqAccess
puts (SCREEN, "Quartz access (less loop overhead):   ") ? QAccess

